@@ -1,33 +1,69 @@
+import math
 from collections import deque
 from heapq import heappush, heappop
-from typing import Optional, Iterable, Set
+from typing import Optional, Set, Union, Callable, Iterable, Hashable, Tuple
 
-import math
-
-from .api import Graph, BellmanFordResult, DistanceTable, DistanceEntry, MstResult
-from ..backend.api import DataBackend
-from ..backend.memory import InMemoryBackend
-from ..components.edge import Edge
-from ..components.node import Node
+from grapresso.components.edge import Edge
+from grapresso.components.node import Node
+from .api import BellmanFordResult, DistanceTable, DistanceEntry, MstResult
+from ..backends.api import DataBackend
+from ..backends.memory import InMemoryBackend
 from ..components.path import CircularTour, TourTracker, Path, Cycle, Flow
 from ..datastruct.disjointset import DefaultDisjointSet
 
 
-class DirectedGraph(Graph):
+class DiGraph:
     def __init__(self, data_backend: DataBackend = None):
         if data_backend is None:
             self._nodes_data = InMemoryBackend()
         else:
             self._nodes_data = data_backend
 
+    @property
+    def backend(self) -> DataBackend:
+        """This property offers direct backend access. Use with care.
+        In the long term, this property needs to be made redundant by offering direct Graph API methods as abstraction.
 
-    def __getitem__(self, item):
-        if isinstance(item, tuple) and len(item) == 2:
-            return self.edge(item[0], item[1])
+        Returns: Backend that has been used during initialisation.
+
+        """
+        return self._nodes_data
+
+    def __getitem__(self, item: Union[Node, Edge, Hashable]) -> Union[Optional[Node], Optional[Edge]]:
+        if isinstance(item, Edge):
+            return self.edge(item.from_node, item.to_node)
         else:
-            return self.node_by_name(item)
+            return self.node(item)
 
-    def node_by_name(self, start_node_name) -> Node:
+    # TODO(kdevo): Untested
+    def from_tuples(self, *args: Union[Hashable, Tuple]):
+        for it in args:
+            if not isinstance(it, tuple):
+                self.add_node(it)
+            else:
+                if 1 >= len(it) >= 4 or len(it) == 3 and not isinstance(it[2], dict):
+                    raise ValueError("Only tuples from 1 to 3 values are allowed, "
+                                     "where (u, v) means edge from u to v and with an optional data dict {} such as "
+                                     "(u, v, {'k': 'val'}) and (u, {'k': 'val'}) means node u with a data dict.")
+                elif len(it) == 2 and isinstance(it[1], dict):
+                    self.add_node(it[0], **it[1])
+                elif len(it) == 3 and isinstance(it[2], dict):
+                    self.add_edge(it[0], it[1], **it[2])
+                else:
+                    self.add_edge(it[0], it[1])
+        return self
+
+    # TODO(kdevo): Untested
+    def copy_to(self, graph):
+        for e in self.edges():
+            graph.add_edge(e.from_node, e.to_node, **e.data)
+        return graph
+
+    def edges(self, from_node=None, to_node=None) -> Iterable[Edge]:
+        return [e for e in self._nodes_data.edges()
+                if (not from_node or e.from_node == from_node) and (not to_node or e.to_node == to_node)]
+
+    def node(self, start_node_name) -> Optional[Node]:
         """Convenience helper: Returns a (random) start_node if it is None.
 
         Args:
@@ -36,9 +72,12 @@ class DirectedGraph(Graph):
         Returns
             start_node if start_node is not None else a random node from the graph.
         """
-        return next(iter(self._nodes_data)) if start_node_name is None else self._nodes_data[start_node_name]
+        try:
+            return next(iter(self._nodes_data)) if start_node_name is None else self._nodes_data[start_node_name]
+        except KeyError:
+            return None
 
-    def free_node_name(self, wished_name: str = "Node"):
+    def free_node_name(self, wished_name: str = "Node") -> str:
         """Returns a free node name based on wished_name using a numeric suffix.
 
         Args:
@@ -47,6 +86,7 @@ class DirectedGraph(Graph):
         Returns:
             wished_name [+ suffix_no]
         """
+
         def new_name(start_name, suffix_no):
             return start_name + str(suffix_no)
 
@@ -70,14 +110,16 @@ class DirectedGraph(Graph):
 
     def add_node(self, node_name, **attributes):
         self._nodes_data.add_node(node_name, **attributes)
+        return self
 
-    def remove_node(self, node):
-        raise NotImplementedError()
-        pass
+    def remove_node(self, node_name):
+        self._nodes_data.remove_node(node_name)
 
     def edge(self, from_node_name, to_node_name) -> Optional[Edge]:
         try:
-            return self._nodes_data[from_node_name].edge(self._nodes_data[to_node_name])
+            node = self[from_node_name]
+            if node:
+                return node.edge(self._nodes_data[to_node_name])
         except KeyError:
             return None
 
@@ -114,7 +156,7 @@ class DirectedGraph(Graph):
                 on_visited_cb(current_node)
         return seen
 
-    def perform_kruskal(self, on_new_edge_cb: callable = None):
+    def perform_kruskal(self, on_new_edge_cb: Callable[[Edge], None] = None) -> float:
         dj_set = DefaultDisjointSet(self._nodes_data)
 
         cost = 0
@@ -129,13 +171,14 @@ class DirectedGraph(Graph):
 
         return cost
 
-    def build_mst(self, initialized_graph) -> MstResult:
-        costs = {'kruskal': self.perform_kruskal, 'prim': self.perform_prim}[self._nodes_data.mst_alg_hint](
+    def build_mst(self, initialized_graph, preferred_algorithm=None) -> MstResult:
+        costs = {'kruskal': self.perform_kruskal, 'prim': self.perform_prim}[
+            preferred_algorithm if preferred_algorithm else self._nodes_data.mst_alg_hint](
             on_new_edge_cb=lambda e: initialized_graph.add_edge(e.from_node.name, e.to_node.name, cost=e.cost)
         )
         return MstResult(costs, initialized_graph)
 
-    def perform_prim(self, start_node_name=None, on_new_edge_cb: callable = None):
+    def perform_prim(self, start_node_name=None, on_new_edge_cb: callable = None) -> int:
         to_visit = set(self._nodes_data)
         current_node = self[start_node_name]
 
@@ -230,7 +273,10 @@ class DirectedGraph(Graph):
         return Path.from_tree(lambda v: bmr.dist_table[v].parent, self[start_node_name], self[end_node_name])
 
     def __str__(self):
-        return "Node count: {} | Edges: {}".format(len(self._nodes_data), *self._nodes_data.edges())
+        return self.__repr__()
+
+    def __repr__(self):
+        return "Directed graph with {} nodes".format(len(self._nodes_data))
 
     def __len__(self):
         return len(self._nodes_data)
@@ -238,7 +284,7 @@ class DirectedGraph(Graph):
     def __sizeof__(self):
         return self._nodes_data.__sizeof__()
 
-    def build_residual_graph(self, initialized_residual_graph: 'DirectedGraph', flow: Flow):
+    def build_residual_graph(self, initialized_residual_graph: 'DiGraph', flow: Flow):
         edge_info = {}
 
         for edge in self._nodes_data.edges():
@@ -289,7 +335,7 @@ class DirectedGraph(Graph):
 
         while path_in_resgraph_exists:
             # Calculate G' and u':
-            res_graph, res_edge_to_orig = self.build_residual_graph(DirectedGraph(InMemoryBackend()), flow)
+            res_graph, res_edge_to_orig = self.build_residual_graph(DiGraph(InMemoryBackend()), flow)
             # Find (s,t)-path p ∈ G' and also therewith determine γ = min(u'(e)) with e ∈ p:
             augmenting_path = res_graph.shortest_path(source_node_name, target_node_name)
             if augmenting_path is None:
@@ -299,7 +345,7 @@ class DirectedGraph(Graph):
                 flow.augment_along_path(augmenting_path, res_edge_to_orig)
         return flow
 
-    def perform_cycle_cancelling(self):
+    def perform_cycle_cancelling(self) -> Flow:
         # Create a virtual SUPER-SOURCE s and SUPER-TARGET t
         s_name, t_name = self.free_node_name("SUPER-SOURCE"), self.free_node_name("SUPER-TARGET")
         s_out_flow, t_in_flow = 0, 0
@@ -327,7 +373,7 @@ class DirectedGraph(Graph):
         # Now that we have a valid b-flow, we can start the iterations until no cost-negative cycle is found anymore:
         while True:
             # Calculate G', u', c':
-            res_graph, res_edge_to_orig = self.build_residual_graph(DirectedGraph(InMemoryBackend()), flow)
+            res_graph, res_edge_to_orig = self.build_residual_graph(DiGraph(InMemoryBackend()), flow)
 
             # We can also find negative cycles when the graph's connected component is bigger than 1.
             # That is why we need to test if there is a negative cycle from all start nodes:
@@ -347,7 +393,7 @@ class DirectedGraph(Graph):
                 # If we cannot find a negative cycle anymore, we found an optimal solution:
                 return flow
 
-    def perform_successive_shortest_path(self):
+    def perform_successive_shortest_path(self) -> Flow:
         pseudo_balances = {}
 
         def pseudo_balances_satisfied():
@@ -372,7 +418,7 @@ class DirectedGraph(Graph):
 
         while not pseudo_balances_satisfied():
             # Calc residual graph G', we need it when checking if the target is accessible from the source:
-            res_graph, res_edge_to_orig = self.build_residual_graph(DirectedGraph(InMemoryBackend()), flow)
+            res_graph, res_edge_to_orig = self.build_residual_graph(DiGraph(InMemoryBackend()), flow)
 
             try:
                 # Find a pseudo source s, where we need to saturate the outflow in G':
@@ -402,7 +448,7 @@ class DirectedGraph(Graph):
         return flow
 
 
-class UndirectedGraph(DirectedGraph):
+class UnDiGraph(DiGraph):
     def add_edge(self, a, b, **kwargs):
         if a not in self._nodes_data:
             self.add_node(a)
@@ -426,15 +472,15 @@ class UndirectedGraph(DirectedGraph):
         with CircularTour(node) as nn_tour:
             while len(to_visit) - 1 > 0:
                 to_visit.remove(node)
-                edge = next(e for e in node.updated_sorted_edges() if e.to_node in to_visit)
+                edge = next(e for e in node.sorted_edges() if e.to_node in to_visit)
                 nn_tour.go(edge)
                 node = edge.to_node
         return nn_tour
 
     def double_tree_tour(self, start_node_name=None):
-        mst_result = self.build_mst(UndirectedGraph(InMemoryBackend()))
+        mst_result = self.build_mst(UnDiGraph(InMemoryBackend()))
         visited_node_names = []
-        mst_result.mst.perform_dfs(start_node_name, on_visited_cb=lambda n: visited_node_names.append(n.name))
+        mst_result.tree.perform_dfs(start_node_name, on_visited_cb=lambda n: visited_node_names.append(n.name))
         with CircularTour(self[start_node_name]) as dt_tour:
             for i in range(0, len(visited_node_names) - 1):
                 for edge in self._nodes_data[visited_node_names[i]].edges:
@@ -510,7 +556,7 @@ class UndirectedGraph(DirectedGraph):
 
     def max_matchings(self, node_names_a: Set, node_names_b: Set):
         # Create a directed graph
-        directed_graph = DirectedGraph(InMemoryBackend())
+        directed_graph = DiGraph(InMemoryBackend())
         for edge in self._nodes_data.edges():
             if edge.from_node.name in node_names_a and edge.to_node.name in node_names_b:
                 directed_graph.add_edge(edge.from_node.name, edge.to_node.name, capacity=1)
@@ -530,3 +576,5 @@ class UndirectedGraph(DirectedGraph):
                 matched_edges.append(self.edge(e.from_node.name, e.to_node.name))
         return matched_edges
 
+    def __repr__(self):
+        return "Undirected graph with {} nodes".format(len(self._nodes_data))
